/* sensors.c	- requires GKrellM 1.2.0 or better
|		gcc -fPIC `gtk-config --cflags` `imlib-config --cflags-gdk` -c sensors.c
|		gcc -shared -Wl -o sensors.so sensors.o
|		gkrellm -p sensors.so
| to install it,
|		cp sensors.so ~/.gkrellm/plugins/
*/

#define UPDATE_DELAY 10
  /*
  | Dfini tous les "combien" on fait la lecture de temperature
  */


#include <gkrellm/gkrellm.h>
#include <locale.h>

#include <sys/types.h>
#include <dirent.h>


  /* CONFIG_NAME would be the name in the configuration tree, but this demo
  |  does not have a config.  See demo1.c for that.
  */
#define	CONFIG_NAME	"LM-Sensors"

  /* STYLE_NAME will be the theme subdirectory for custom images for this
  |  plugin and it will be the gkrellmrc style name for custom settings.
  */
#define	STYLE_NAME	"sensors-style"

static Monitor	*monitor;
static Panel	*panel = NULL ;
static GtkWidget	*panelbox = NULL ;

static char sensor_path[512] = "/proc/sys/dev/sensors" ;

static gint sensor_unit = 0 ; // ==0 -> C , <>0 -> F

typedef struct _SensorItem SensorItem ;
struct _SensorItem {
	char* name;
	char* format;
} ;

static GList *sensors = NULL;
static GList *decals = NULL;
static gint	style_id;

static GList *sensors_entries = NULL;

static GtkWidget	*sensor_unit_check ;

void FreeSensorItem(gpointer data, gpointer user_data) {
//	printf("FreeSensorItem [%s]\n",((SensorItem*)data)->name) ;
	free(((SensorItem*)data)->name) ;
	free(((SensorItem*)data)->format) ;
	free(data) ;
}

void FreeDecal(gpointer data, gpointer user_data) {
//	printf("FreeDecal\n") ;
	gkrellm_remove_and_destroy_decal(panel,(Decal*)data) ;
}


SensorItem* NewSensorItem (char *name, char *format) {
	SensorItem *result ;
	result = malloc(sizeof(SensorItem)) ;
	result->name = strdup(name) ;
	result->format = strdup(format) ;
	return result ;
}

static gint
panel_expose_event(GtkWidget *widget, GdkEventExpose *ev)
	{
	gdk_draw_pixmap(widget->window,
			widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			panel->pixmap, ev->area.x, ev->area.y, ev->area.x, ev->area.y,
			ev->area.width, ev->area.height);

	return FALSE;
	}

inline float Temperature(float temp) {
	return (sensor_unit?((temp * 1.8) + 32.0):temp) ;
}

typedef struct {
	gint		height ;
	Style		*style ;
	TextStyle	*textstyle ;
} CreateDecalContext;

void CreateDecal (gpointer data, gpointer user_data) {
	CreateDecalContext	*context = user_data ;
	Decal			*decal ;
	SensorItem *sensoritem = data ;

//	printf("CreateDecal [%s]::[%s]\n",sensoritem->name,sensoritem->format);

	decal = gkrellm_create_decal_text(panel, "Ay", context->textstyle, context->style, -1, context->height , -1);
	context->height = decal->y + decal->h ;

	if (decals) {
		g_list_append(decals, decal) ;
	} else {
		decals = g_list_alloc() ;
		decals->data = decal ;
	}

}

void CreateDecals(gint first_create) {
	Style *style ;
	CreateDecalContext context ;

//	printf("CreateDecals %i\n",first_create) ;

	// les decals sont supprims !
	decals = NULL ;

	if (first_create) {
		panel = gkrellm_panel_new0();
	}


	context.height = -1 ;
	context.style = gkrellm_meter_style(style_id);
	context.textstyle = gkrellm_meter_textstyle(style_id);
	g_list_foreach(sensors,CreateDecal,&context) ;

	style = gkrellm_meter_style(style_id);
	gkrellm_panel_configure(panel, NULL, style);
	gkrellm_panel_create(panelbox, monitor, panel);

	//gkrellm_monitor_height_adjust(-panel->h);

	if (first_create)
	    gtk_signal_connect(GTK_OBJECT (panel->drawing_area), "expose_event",
    	        (GtkSignalFunc) panel_expose_event, NULL);

//	printf("good!\n") ;

}

void format_sensor(gchar * name, gchar * format, gint pos, gchar * dest, gint max) {
	gchar line[512], cpy[512], * sub ;
	FILE * fp ;
	gint i ;
	gfloat val ;

	strncpy(line, sensor_path, 512) ;
	strncat(line, "/", 512) ;
	strncat(line, name, 512) ;

	if ((fp = fopen(line,"r")) != NULL) {
		fgets(line, 512, fp) ;
		fclose(fp) ;

		// When LOCALE use ',' as decimal point (for example)
		// we must translate '.' from the sensor output to ','
		if (localeconv()->decimal_point[0] != '.') {
			while ((sub = strstr(line, "."))) {
				sub[0] = localeconv()->decimal_point[0] ;
			} // while
		} // if

		for (i = 0 ; i < pos ; i++) {
			sscanf(line, "%f %[^\n]", &val, cpy) ;
			strncpy(line, cpy, 512) ;
		}

		if ((strstr(name,"temp")!=NULL)) val = Temperature(val) ;

		snprintf(dest, max, format, val) ;
	} else {
		strcat(dest,"") ;
	}

}

gint sensor_pos(gchar *name) {
	if ((strstr(name,"in") != NULL) || (strstr(name,"temp") != NULL)) {
		return 3 ;
	} else {
		if ((strstr(name,"fan") != NULL)) {
			return 2 ;
		} else {
			return 1 ;
		}
	}
}

void UpdateDecal (gpointer data, gpointer user_data) {
	Decal *decal = data ;
	gint *pos = user_data ;
	SensorItem *sensoritem ;
	gchar line[512] ;

//	printf("\tstart UpdateDecal s:%i d:%i\n",g_list_length(sensors),g_list_length(decals));

	decal->x_off = 2 ;
	sensoritem = (SensorItem*)(g_list_nth(sensors,*pos)->data) ;
	format_sensor(sensoritem->name, sensoritem->format,sensor_pos(sensoritem->name),line, 512) ;
//	snprintf(line, 512, sensoritem->format,32.5) ;
	gkrellm_draw_decal_text(panel, decal , line , GK.timer_ticks);

	// next sensor, walk throw decals AND sensors
	(*pos)++ ;

//	printf("\tend UpdateDecal\n");

}

static void
update_plugin()
{
	gint pos ;

	if ((GK.timer_ticks % UPDATE_DELAY)==1) { // We update all UPDATE_DELAY
//		printf("update_plugin\n");
		pos = 0 ;
		g_list_foreach(decals, UpdateDecal, &pos) ;
		gkrellm_draw_panel_layers(panel);
	}

} // update_plugin


static void
create_plugin(GtkWidget *vbox, gint first_create) {

//	printf("start create_plugin\n");

	if (panelbox == NULL) {
		panelbox = gtk_vbox_new(FALSE,0) ;
		gtk_container_add(GTK_CONTAINER(vbox), panelbox) ;
		gtk_widget_show(panelbox) ;
	}

	CreateDecals(first_create) ;
//	printf("end create_plugin\n");

}


/* Configuration */
#define PLUGIN_CONFIG_KEYWORD	"sensors-cfg"
#define PLUGIN_CONFIG_UNIT "UNIT"

void SaveEntry (gpointer data, gpointer user_data) {
	GtkWidget *entry = data;
	gchar *name ;
	gchar *text ;

	name = gtk_widget_get_name(entry) ;
	text = gtk_entry_get_text(GTK_ENTRY(entry)) ;
	if (strlen(text)>0) {
		if (user_data) {
			fprintf(user_data, "%s %s %s\n", PLUGIN_CONFIG_KEYWORD, name, text);
//			printf("Save->[%s]::[%s]\n", name, text) ;
		}
	}
}

void SynchronizeEntry (gpointer data, gpointer user_data) {
	GtkWidget *entry = data;
	gchar *name ;
	gchar *text ;
	SensorItem* item ;

	name = gtk_widget_get_name(entry) ;
	text = gtk_entry_get_text(GTK_ENTRY(entry)) ;
	if (strlen(text)>0) {
//		printf("AddSensor->[%s]::[%s]\n", name, text) ;
		item = NewSensorItem(name, text) ;
		if (sensors) {
			g_list_append(sensors, item) ;
		} else {
			sensors = g_list_alloc() ;
			sensors->data = item ;
		}
	}
}

void SaveEntries(FILE* f) {
//	printf("SaveEntries\n");
	g_list_foreach(sensors_entries,&SaveEntry,f) ;

}

void SynchronizeEntries() {
//	printf("SynchronizeEntries\n");
	g_list_foreach(sensors, FreeSensorItem,NULL) ;
	g_list_free(sensors) ;
	sensors = NULL ;
	g_list_foreach(sensors_entries,&SynchronizeEntry,NULL) ;

}

static void
save_plugin_config(FILE *f) {
//	printf("save_plugin_config\n");
	fprintf(f, "%s %s %i\n", PLUGIN_CONFIG_KEYWORD, PLUGIN_CONFIG_UNIT, sensor_unit);
//	printf("Save->[%s]::[%i]\n", PLUGIN_CONFIG_UNIT, sensor_unit) ;
	SaveEntries(f) ;
//	printf("All is Good\n") ;
}

static void
load_plugin_config(gchar *arg)
{
	gchar	config[64], item[256], chaine[256];
	gint	n;
	SensorItem * sensoritem ;

//	printf("load_plugin_config\n");

	n = sscanf(arg, "%s %[^\n]", config, item);
	if (n == 2)
	{
		if (strcmp(config, PLUGIN_CONFIG_UNIT) == 0) {
			sscanf(item, "%i\n", &sensor_unit);
//			printf("Read->[%s]::[%d]\n", config, sensor_unit) ;
		} else {
			sscanf(item, "%[^\n]", chaine) ;
//			printf("Read->[%s]::[%s]\n", config, chaine) ;
			sensoritem = NewSensorItem(config, chaine) ;
			if (sensors) {
				g_list_append(sensors, sensoritem) ;
			} else {
				sensors = g_list_alloc() ;
				sensors->data = sensoritem ;
			}
		}
	}
}

static void
apply_plugin_config() {
	CreateDecalContext context ;
	gint h ;
//	printf("apply_plugin_config\n");

	sensor_unit = GTK_TOGGLE_BUTTON(sensor_unit_check)->active ;
	// d'abord dtruire tous les affichages :

	g_list_foreach(decals, FreeDecal,NULL) ;
	decals = NULL ;
	h = panel->h ;
	gkrellm_destroy_panel(panel) ;
	panel = NULL ;
	gkrellm_monitor_height_adjust(-h);


	// puis crer les nouveaux :
	SynchronizeEntries() ;
	CreateDecals(1) ;

}

static gchar	*plugin_info_text[] =
{
_("<b>LM-Sensors\n"),
_("\nThis plugin shows all sensors you've got a driver for.\n\n"),
_("<b>Required :\n"),
_("You must install lm-sensors to use this plugins. Find if at : \n"),
"http://www.netroedge.com/~lm78/download.html\n\n",
_("<b>Usage :\n"),
_("<i>in*, fan*, temp*"),_(" : if you want to see this sensor, juste enter a formating string like : '%.1f' or 'Tcpu%.1fC' or 'Fan1:%.0f rpm'.\n"),
"\n",
_("<i>convert C to F"),_(" : check it to convert to F.\n")
};

static gchar *about_text =
_(
"LM-Sensors Monitor\n"
"Copyright (C) 2001 Frederic Bonnaud\n"
"frederic.bonnaud@laposte.net\n"
"http://encore.inconnue.net\n\n"
"Released under the GNU Public Licence"
);

void free_data (gpointer data, gpointer user_data) {
	free(data) ;
}

typedef struct {
	void *p1 ;
	void *p2 ;

} DOUBLE_PTR ;

gint is_name(gconstpointer a, gconstpointer b) {
	const SensorItem *A = a ;
	const SensorItem *B = b ;
	//printf("compare (%s,%s) = %i\n", A->name, B->name, strcmp(A->name, B->name))  ;
	return (strcmp(A->name, B->name)) ;
}


char* FindFormat(char* filename) {
	SensorItem sensoritem ;
	GList *result ;

	sensoritem.name = filename ;
	result = g_list_find_custom(sensors, &sensoritem, is_name) ;
	return (result ? ((SensorItem*)(result->data))-> format : "" ) ;
}

void Create_SensorItem (gpointer data, gpointer user_data) {
	char *filename = (char*)data ;
	char tampon[256] ;
	GtkWidget *vbox = ((DOUBLE_PTR*)user_data)->p1 ;
	char *dirname = ((DOUBLE_PTR*)user_data)->p2 ;

	GtkWidget *sensor_box, *sensor_label, *sensor_entry ;

	sensor_box = gtk_hbox_new (FALSE, 0);
	snprintf(tampon,256,"sensor_box_%s/%s",dirname,filename) ;
	gtk_widget_set_name (sensor_box, tampon);
	gtk_widget_ref (sensor_box);
	gtk_object_set_data_full (GTK_OBJECT (vbox), tampon, sensor_box,
																	(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (sensor_box);
	gtk_box_pack_start (GTK_BOX (vbox), sensor_box, TRUE, TRUE, 0);

	snprintf(tampon,256,"%s : ",filename) ;
	sensor_label = gtk_label_new (tampon);
	snprintf(tampon,256,"sensor_label_%s/%s",dirname,filename) ;
	gtk_widget_set_name (sensor_label, tampon);
	gtk_widget_ref (sensor_label);
	gtk_object_set_data_full (GTK_OBJECT (vbox), tampon, sensor_label,
																	(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (sensor_label);
	gtk_box_pack_start (GTK_BOX (sensor_box), sensor_label, FALSE, FALSE, 0);

	snprintf(tampon,256,"%s/%s",dirname,filename) ;
	sensor_entry = gtk_entry_new ();
	gtk_entry_set_text(GTK_ENTRY(sensor_entry), FindFormat(tampon));
	gtk_widget_set_name (sensor_entry, tampon);
	gtk_widget_ref (sensor_entry);
	gtk_object_set_data_full (GTK_OBJECT (vbox), tampon, sensor_entry,
																	(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (sensor_entry);
	gtk_box_pack_start (GTK_BOX (sensor_box), sensor_entry, TRUE, TRUE, 0);

	if (sensors_entries) {
		g_list_append(sensors_entries, sensor_entry) ;
	} else {
		sensors_entries = g_list_alloc() ;
		sensors_entries->data = sensor_entry ;
	}

}

void Create_SensorPage (gpointer data, gpointer user_data) {
	DIR * sensors_dir ;
	struct dirent * entry ;
	GtkWidget *vbox ;
	DOUBLE_PTR double_ptr ;
	char path[256] ;
	GList *liste ;

	if (data) {

		strncpy(path, sensor_path,256) ;
		strncat(path, "/", 256) ;
		strncat(path, (char*)data, 256) ;
		sensors_dir = opendir(path) ;

		liste = NULL ;

		while ((entry = readdir(sensors_dir))) {
			if (
			// it's not a real test, we'd test : in\d* etc..
					((strstr(entry->d_name, "in")!=NULL) ||
					 (strstr(entry->d_name, "fan")!=NULL) ||
					 (strstr(entry->d_name, "temp")!=NULL))
					&& (strstr(entry->d_name, "_div")==NULL)
				) {
				if (liste) {
					g_list_append(liste, strdup(entry->d_name)) ;
				} else {
					liste = g_list_alloc() ;
					liste->data = strdup(entry->d_name) ;
				} // if liste
			} else {
			} // if is sensor
		} // while entry
		closedir(sensors_dir) ;

		if (liste != NULL) {
			vbox = gkrellm_create_framed_tab((GtkWidget*)user_data, (char*)data);
			double_ptr.p1 = vbox ;
			double_ptr.p2 = data ;
			g_list_foreach(liste, &Create_SensorItem, &double_ptr) ;
		} // if liste

		// nettoyage !
		g_list_foreach(liste,&free_data, NULL) ;
		g_list_free(liste) ;
	} else {
		printf("warning : data == NULL in Create_SensorPage\n") ;
	} // if data
}

static void dir_read(GtkWidget *tabs) {

	FILE* fchips ;
	char chips[256], line[256], chip[256] ;
	int i ;
	GList * liste ;
	GtkWidget* notebook, *vbox ;

	if (sensors_entries != NULL) {
		g_list_free(sensors_entries) ;
		sensors_entries = NULL ;
	} // if sensors_entries

	strncpy(chips,sensor_path, 256) ;
	strncat(chips,"/chips", 256) ;
	fchips = fopen(chips,"r") ;

	liste = NULL ;


	if (fchips) {
		while (fgets(line, 256, fchips) ) {
			sscanf(line,"%i %s\n", &i, chip) ;
			if (liste) {
				g_list_append(liste, strdup(chip)) ;
			} else {
				liste = g_list_alloc() ;
				liste->data = strdup(chip) ;
			} // if liste
		} // if line
	} // if fchips

	vbox = gkrellm_create_framed_tab(tabs, _("Chipsets"));
	notebook = gtk_notebook_new ();

	gtk_widget_set_name (notebook, "notebook");
        gtk_widget_ref (notebook);
        gtk_object_set_data_full (GTK_OBJECT (vbox), "notebook", notebook,
                                                        (GtkDestroyNotify) gtk_widget_unref);
	//gtk_notebook_set_tab_pos (GTK_NOTEBOOK(notebook), GTK_POS_LEFT);

        gtk_widget_show (notebook);
        gtk_box_pack_start (GTK_BOX (vbox), notebook, TRUE, TRUE, 0);


	g_list_foreach(liste,&Create_SensorPage, notebook) ;


	// nettoyage !
	g_list_foreach(liste,&free_data, NULL) ;
	g_list_free(liste) ;
}

static void
create_plugin_tab(GtkWidget *tab_vbox)
	{
	GtkWidget		*tabs;
	GtkWidget		*vbox;
	GtkWidget		*text;
	GtkWidget		*about;
	GtkWidget		*unit_box ;

	gint				i;

	/* Make a couple of tabs.  One for setup and one for info
	*/
	tabs = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(tabs), GTK_POS_TOP);
	gtk_box_pack_start(GTK_BOX(tab_vbox), tabs, TRUE, TRUE, 0);

/* --Chipsets tab */
	dir_read(tabs) ;


	// unit selection
		vbox = gkrellm_create_framed_tab(tabs, _("Misc..."));
        unit_box = gtk_hbox_new (FALSE, 0);
        gtk_widget_set_name (unit_box, "unit_box");
        gtk_widget_ref (unit_box);
        gtk_object_set_data_full (GTK_OBJECT (vbox), "unit_box", unit_box,
                                                        (GtkDestroyNotify) gtk_widget_unref);
        gtk_widget_show (unit_box);
        gtk_box_pack_start (GTK_BOX (vbox), unit_box, FALSE, TRUE, 0);

        sensor_unit_check = gtk_check_button_new_with_label (_(" convert to F"));
        gtk_widget_set_name (sensor_unit_check, "sensor_unit_check");
        gtk_widget_ref (sensor_unit_check);
        gtk_object_set_data_full (GTK_OBJECT (vbox), "sensor_unit_check", sensor_unit_check,
                                                        (GtkDestroyNotify) gtk_widget_unref);
	GTK_TOGGLE_BUTTON(sensor_unit_check)->active = (gboolean)sensor_unit ;
        gtk_widget_show (sensor_unit_check);
        gtk_box_pack_start (GTK_BOX (unit_box), sensor_unit_check, FALSE, TRUE, 0);
        gtk_widget_set_usize (sensor_unit_check, 220, -2);

/* --Info tab */
	vbox = gkrellm_create_framed_tab(tabs, _("Info"));
	text = gkrellm_scrolled_text(vbox, NULL,
				GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	for (i = 0; i < sizeof(plugin_info_text)/sizeof(gchar *); ++i)
		gkrellm_add_info_text_string(text, plugin_info_text[i]);

/* -- About tab */
	vbox = gkrellm_create_framed_tab(tabs, _("About"));
	about = gtk_label_new(about_text) ;
	gtk_box_pack_start(GTK_BOX(vbox), about, TRUE, TRUE, 0);


	}

/* The monitor structure tells GKrellM how to call the plugin routines.
*/
static Monitor	plugin_mon	=
	{
	CONFIG_NAME,				/* Name, for config tab.    */
	0,								/* Id,  0 if a plugin       */
	create_plugin,				/* The create function      */
	update_plugin,				/* The update function      */
	create_plugin_tab,		/* The config tab create function   */
	apply_plugin_config, 	/* Apply the config function        */

	save_plugin_config,		/* Save user config			*/
	load_plugin_config,		/* Load user config			*/
	PLUGIN_CONFIG_KEYWORD,	/* config keyword			*/

	NULL,				/* Undefined 2	*/
	NULL,				/* Undefined 1	*/
	NULL,				/* Undefined 0	*/

	MON_CPU,			/* Insert plugin before this monitor.  Choose	*/
							/*   MON_CLOCK, MON_CPU, MON_PROC, MON_DISK,	*/
							/*   MON_INET, MON_NET, MON_FS, MON_MAIL,		*/
							/*   MON_APM, or MON_UPTIME						*/
							/* Choose wisely, others have to live with it.	*/

	NULL,					/* Handle if a plugin, filled in by GKrellM     */
	NULL					/* path if a plugin, filled in by GKrellM       */
	};

  /* All GKrellM plugins must have one global routine named init_plugin()
  |  which returns a pointer to a filled in monitor structure.
  */
Monitor *
init_plugin()
	{
	/* If this call is made, the background and krell images for this plugin
	|  can be custom themed by putting bg_meter.png or krell.png in the
	|  subdirectory STYLE_NAME of the theme directory.  Text colors (and
	|  other things) can also be specified for the plugin with gkrellmrc
	|  lines like:  StyleMeter  STYLE_NAME.textcolor orange black shadow
	|  If no custom themeing has been done, then all above calls using
	|  style_id will be equivalent to style_id = DEFAULT_STYLE_ID.
	*/
	style_id = gkrellm_add_meter_style(&plugin_mon, STYLE_NAME);
	monitor = &plugin_mon;
	return &plugin_mon;
	}
